/*

Copyright (c) 2004 Paul Hsieh
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer. 

    Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution. 

    Neither the name of LL nor the names of its contributors may be used 
    to endorse or promote products derived from this software without 
    specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

*/

#include "ll.h"

#ifdef MEMORY_DEBUG
#include "memdbg.h"
#endif

struct genLinkedListNode {
	void * data;
	struct genLinkedListNode * next;
};

/*
 *	generic list abstract data type.
 */
int utilGenLinkedListInit (struct genLinkedList * gl) {
	if (gl == NULL) return -__LINE__;
	gl->head = gl->last = NULL;
	return 0;
}

int utilGenLinkedListUnInit (struct genLinkedList * gl) {
struct genLinkedListNode * node1, * node2;
	if (gl == NULL) return -__LINE__;
	node1 = gl->head;
	while (node1) {
		node2 = node1->next;
		node1->next = NULL;
		free (node1);
		node1 = node2;
	}
	gl->head = gl->last = NULL;
	return 0;
}

int utilGenLinkedListAppend (struct genLinkedList * gl, void * e) {
struct genLinkedListNode * node;
	if (gl == NULL) return -__LINE__;

	if (gl->head == NULL) {
		if (gl->last != NULL) return -__LINE__;
		if (NULL == (node = (struct genLinkedListNode *) malloc (sizeof (struct genLinkedListNode)))) return -__LINE__;
		node->data = e;
		node->next = NULL;
		gl->head = gl->last = node;
		return 0;
	}

	if (gl->last == NULL) return -__LINE__;
	if (NULL == (node = (struct genLinkedListNode *) malloc (sizeof (struct genLinkedListNode)))) return -__LINE__;
	node->data = e;
	node->next = NULL;

	gl->last->next = node;
	gl->last = node;
	return 0;
}

int utilGenLinkedListPush (struct genLinkedList * gl, void * e) {
struct genLinkedListNode * node;
	if (gl == NULL) return -__LINE__;

	if (gl->head == NULL) {
		if (gl->last != NULL) return -__LINE__;
		if (NULL == (node = (struct genLinkedListNode *) malloc (sizeof (struct genLinkedListNode)))) return -__LINE__;
		node->data = e;
		node->next = NULL;
		gl->head = gl->last = node;
		return 0;
	}

	if (gl->last == NULL) return -__LINE__;
	if (NULL == (node = (struct genLinkedListNode *) malloc (sizeof (struct genLinkedListNode)))) return -__LINE__;
	node->data = e;
	node->next = gl->head;
	gl->head = node;
	return 0;
}

int utilGenLinkedListPop (struct genLinkedList * gl, void ** e) {
struct genLinkedListNode * node;
	if (gl == NULL) return -__LINE__;
	if (NULL == (node = gl->head)) return -__LINE__;
	gl->head = node->next;
	if (e) *e = node->data;
	node->next = NULL;
	free (node);
	return 0;
}

int utilGenLinkedListIteratorStart (struct genLinkedList * gl, struct genLinkedListIterator * iter) {
	if (gl == NULL || iter == NULL) return -__LINE__;
	iter->curr = gl->head;
	return 0;
}

int utilGenLinkedListIteratorNext (struct genLinkedListIterator * iter) {
	if (iter == NULL) return -__LINE__;
	iter->curr = (NULL == iter->curr) ? NULL : iter->curr->next;
	return NULL != iter->curr;
}

int utilGenLinkedListIteratorEndReached (struct genLinkedListIterator * iter) {
	if (iter == NULL) return -__LINE__;
	return NULL == iter->curr;
}

int utilGenLinkedListIteratorGet (struct genLinkedListIterator * iter, void ** out) {
	if (iter == NULL || out == NULL) return -__LINE__;
	if (NULL == iter->curr) return -__LINE__;
	*out = iter->curr->data;
	return 0;
}

int utilGenLinkedListIteratorInsertAfter (struct genLinkedListIterator * iter, void * e) {
struct genLinkedListNode * node;
	if (iter == NULL) return -__LINE__;
	if (NULL == iter->curr) return -__LINE__;
	if (NULL == (node = (struct genLinkedListNode *) malloc (sizeof (struct genLinkedListNode)))) return -__LINE__;

	node->next = iter->curr->next;
	node->data = e;
	iter->curr->next = node;
	return 0;
}

int utilGenLinkedListIteratorRemoveAfter (struct genLinkedListIterator * iter, void ** e) {
struct genLinkedListNode * node;
	if (iter == NULL) return -__LINE__;
	if (NULL == iter->curr) return -__LINE__;
	node = iter->curr->next;
	if (NULL == node) return -1;
	if (e) *e = node->data;
	iter->curr->next = node->next;
	node->data = NULL;
	node->next = NULL;
	free (node);
	return 0;
}

