/*

Copyright (c) 2004 Paul Hsieh
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer. 

    Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution. 

    Neither the name of LL nor the names of its contributors may be used 
    to endorse or promote products derived from this software without 
    specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef LL_INCLUDE
#define LL_INCLUDE

#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 *  Generic linked list on void * type.  Just declare a struct genLinkedList 
 *  and operate on it directly with the API functions given below.
 */

struct genLinkedList {
	struct genLinkedListNode * head, * last;
};

struct genLinkedListIterator {
	struct genLinkedListNode * curr;
};

extern int utilGenLinkedListInit (struct genLinkedList * gl);
extern int utilGenLinkedListUnInit (struct genLinkedList * gl);
extern int utilGenLinkedListAppend (struct genLinkedList * gl, void * e);
extern int utilGenLinkedListPush (struct genLinkedList * gl, void * e);
extern int utilGenLinkedListPop (struct genLinkedList * gl, void ** e);
extern int utilGenLinkedListIteratorStart (struct genLinkedList * gl, struct genLinkedListIterator * iter);
extern int utilGenLinkedListIteratorNext (struct genLinkedListIterator * iter);
extern int utilGenLinkedListIteratorEndReached (struct genLinkedListIterator * iter);
extern int utilGenLinkedListIteratorGet (struct genLinkedListIterator * iter, void ** out);
extern int utilGenLinkedListIteratorInsertAfter (struct genLinkedListIterator * iter, void * e);
extern int utilGenLinkedListIteratorRemoveAfter (struct genLinkedListIterator * iter, void ** e);

#define llo__glue2(x,y) x##y
#define llo__glue3(x,y,z) x##y##z

/*
 *  Specific opaque linked list for a specific pointer type.  Getting the API 
 *  just requires invoking llTemplatePrototypes(ptype), and proving the 
 *  implemetation just requires invoking llTemplateImpl(ptype).
 */

#define llOf_Empty {{NULL, NULL}}
#define llOf_Iter_Empty {{NULL}}

#define llTemplateType(ptype) \
typedef struct llo__glue2(tagLinkedListOf_,ptype)     llo__glue2(llOf_,ptype); \
typedef struct llo__glue2(tagLinkedListOfIter_,ptype) llo__glue2(llOf_iter_,ptype); \
struct llo__glue2(tagLinkedListOfIter_,ptype) { \
    struct genLinkedListIterator i; \
};

#define llTemplatePrototypes(ptype) \
llTemplateType (ptype) \
extern llo__glue2(llOf_,ptype) * llo__glue3(llo_,ptype,_new) (void); \
extern int llo__glue3(llo_,ptype,_destroy)             (llo__glue2(llOf_,ptype) * g);\
extern int llo__glue3(llo_,ptype,_init)                (llo__glue2(llOf_,ptype) * g);\
extern int llo__glue3(llo_,ptype,_uninit)              (llo__glue2(llOf_,ptype) * g);\
extern int llo__glue3(llo_,ptype,_append)              (llo__glue2(llOf_,ptype) * g, ptype e);\
extern int llo__glue3(llo_,ptype,_push)                (llo__glue2(llOf_,ptype) * g, ptype e);\
extern int llo__glue3(llo_,ptype,_pop)                 (llo__glue2(llOf_,ptype) * g, ptype * e);\
extern int llo__glue3(llo_,ptype,_iterate_start)       (llo__glue2(llOf_,ptype) * g, llo__glue2(llOf_iter_,ptype) * c);\
extern int llo__glue3(llo_,ptype,_iterate_next)        (llo__glue2(llOf_iter_,ptype) * c);\
extern int llo__glue3(llo_,ptype,_iterate_isend)       (llo__glue2(llOf_iter_,ptype) * c);\
extern int llo__glue3(llo_,ptype,_iterate_get)         (llo__glue2(llOf_iter_,ptype) * c, ptype * e);\
extern int llo__glue3(llo_,ptype,_iterate_insertafter) (llo__glue2(llOf_iter_,ptype) * c, ptype e);\
extern int llo__glue3(llo_,ptype,_iterate_removeafter) (llo__glue2(llOf_iter_,ptype) * c, ptype * e);

#define llTemplateImpl(ptype) \
struct llo__glue2(tagLinkedListOf_,ptype) { \
    struct genLinkedList l; \
}; \
llTemplateType(ptype) \
int llo__glue3(llo_,ptype,_init) (llo__glue2(llOf_,ptype) * g) {\
    return g?utilGenLinkedListInit (&g->l):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_uninit) (llo__glue2(llOf_,ptype) * g) {\
    return g?utilGenLinkedListUnInit (&g->l):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_append) (llo__glue2(llOf_,ptype) * g, ptype e) {\
    return g?utilGenLinkedListAppend (&g->l, (void *) e):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_push) (llo__glue2(llOf_,ptype) * g, ptype e) {\
    return g?utilGenLinkedListPush (&g->l, (void *) e):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_pop) (llo__glue2(llOf_,ptype) * g, ptype * e) {\
    return g?utilGenLinkedListPop (&g->l, (void **) e):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_iterate_start) (llo__glue2(llOf_,ptype) * g, llo__glue2(llOf_iter_,ptype) * c) {\
    return g && c?utilGenLinkedListIteratorStart (&g->l, &c->i):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_iterate_next) (llo__glue2(llOf_iter_,ptype) * c) {\
    return c?utilGenLinkedListIteratorNext (&c->i):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_iterate_isend) (llo__glue2(llOf_iter_,ptype) * c) {\
    return c?utilGenLinkedListIteratorEndReached (&c->i):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_iterate_get) (llo__glue2(llOf_iter_,ptype) * c, ptype * e) {\
    return c?utilGenLinkedListIteratorGet (&c->i, (void **) e):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_iterate_insertafter) (llo__glue2(llOf_iter_,ptype) * c, ptype e) {\
    return c?utilGenLinkedListIteratorInsertAfter (&c->i, (void *) e):-__LINE__;\
}\
int llo__glue3(llo_,ptype,_iterate_removeafter) (llo__glue2(llOf_iter_,ptype) * c, ptype * e) {\
    return c?utilGenLinkedListIteratorRemoveAfter (&c->i, (void **) e):-__LINE__;\
}\
llo__glue2(llOf_,ptype) * llo__glue3(llo_,ptype,_new) (void) {\
llo__glue2(llOf_,ptype) * g = (llo__glue2(llOf_,ptype) *) malloc (sizeof (llo__glue2(llOf_,ptype)));\
    if (g && utilGenLinkedListInit (&g->l) < 0) {\
        free (g);\
        g = (llo__glue2(llOf_,ptype) *)NULL;\
    }\
    return g;\
}\
int llo__glue3(llo_,ptype,_destroy) (llo__glue2(llOf_,ptype) * g) {\
int llo__ret = llo__glue3(llo_,ptype,_uninit) (g);\
    if (llo__ret >= 0) free (g);\
    return llo__ret;\
}

#ifdef __cplusplus
}
#endif

#endif
