#include <stdio.h>
#include <stdlib.h>
#include "ll.h"

struct biography {
	char * name;
	int age;
	char * extra;
};

/* Need to perform a typedef which is a pointer to the data that will be 
   contained to give a single symbol name for type that will be used in the
   API, and correspond to this actual type. */

typedef struct biography * pbio;

#if 0
/* Prototypes for linked list code
   -------------------------------

   Declare the function prototypes for the llO API which includes:

extern llOf_pbio * llo_pbio_new (void);
extern int llo_pbio_destroy             (llOf_pbio * g);
extern int llo_pbio_append              (llOf_pbio * g, pbio e);
extern int llo_pbio_push                (llOf_pbio * g, pbio e);
extern int llo_pbio_pop                 (llOf_pbio * g, pbio * e);
extern int llo_pbio_iterate_start       (llOf_pbio * g, llOf_iter_pbio * c);
extern int llo_pbio_iterate_next        (llOf_iter_pbio * c);
extern int llo_pbio_iterate_isend       (llOf_iter_pbio * c);
extern int llo_pbio_iterate_get         (llOf_iter_pbio * c, pbio * e);
extern int llo_pbio_iterate_insertafter (llOf_iter_pbio * c, pbio e);
extern int llo_pbio_iterate_removeafter (llOf_iter_pbio * c, pbio * e);

   Use this macro in modules which use this linked list API, but which don't
   necessarily contain the implementation.

 */

llTemplatePrototypes (pbio)

#else
/* Generate the linked list code
   -----------------------------

   This will inject the code for llo_pbio_* routines.  This will make a
   linked list API for the type llOf_pbio.  The functions are as 
   described in the prototypes above.  Including the prototypes is not 
   necessary if the implementation is being generated.

   Use this macro exactly once in your project to provide for the 
   implementation of the linked list API.

 */

llTemplateImpl (pbio)

#endif

int main () {
	llOf_pbio * list;
	llOf_iter_pbio iter;

	struct biography testData[3];

	testData[0].name  = "Paul Hsieh";
	testData[0].age   = 34;
	testData[0].extra = "Eccentric";

	testData[1].name  = "Richard Clarke";
	testData[1].age   = 53;
	testData[1].extra = "Patriot";

	testData[2].name  = "Bill Gates";
	testData[2].age   = 51;
	testData[2].extra = "Copycat";

	/* Get a new linked list */
	list = llo_pbio_new ();

	/* Start putting things into the linked list */
	llo_pbio_append (list, &testData[0]);
	llo_pbio_append (list, &testData[1]);
	llo_pbio_append (list, &testData[2]);
	llo_pbio_push   (list, &testData[1]);  /* An alias */

	llo_pbio_iterate_start (list, &iter);
	llo_pbio_iterate_insertafter (&iter, &testData[2]);
	llo_pbio_iterate_removeafter (&iter, NULL);

	/* Iterate over all entries */
	for (llo_pbio_iterate_start (list, &iter);
	     !llo_pbio_iterate_isend (&iter);
	     llo_pbio_iterate_next (&iter)) {
		pbio e;
		llo_pbio_iterate_get (&iter, &e);
		printf ("<%s> <%d> <%s>\n", e->name, e->age, e->extra);
	}

	/* At this point we could iterate over all entries and "destroy" each 
	   one if they were "created". */

	/* Wipe out the linked list container. */
	llo_pbio_destroy (list);

	return 0;
}
